from typing import Type

from asgiref.sync import async_to_sync, sync_to_async
from channels.layers import get_channel_layer
from django.db import models

from rest_live import ListenerEntry, __model_to_listeners
from rest_live.consumers import get_group_name


CREATED = "CREATED"
UPDATED = "UPDATED"
DELETED = "DELETED"


async def send_model_update(
    model: Type[models.Model], instance: models.Model, action: str
):
    model_label = model._meta.label

    listeners: ListenerEntry = __model_to_listeners.get(model_label, dict())
    for group_key_prop, (serializer_class, _) in listeners.items():
        if serializer_class is None:
            return

        group_key = getattr(instance, group_key_prop)
        serializer = serializer_class(instance)
        channel_layer = get_channel_layer()
        group_name = get_group_name(model_label, group_key, group_key_prop)

        @sync_to_async
        def get_serializer_data():
            return serializer.data

        serializer_data = await get_serializer_data()

        content = {"model": model_label, "instance": serializer_data, "action": action, "group_key_value": group_key}

        await channel_layer.group_send(
            group_name,
            {
                "type": "notify",
                "content": content,
                "group_key": group_key_prop,
                "instance_pk": instance.pk,
            },
        )


def onsave_callback(
    sender: Type[models.Model], instance: models.Model, created: bool, **kwargs
):
    async_to_sync(send_model_update)(sender, instance, CREATED if created else UPDATED)


def ondelete_callback(sender, instance, **kwargs):
    async_to_sync(send_model_update)(sender, instance, DELETED)



from typing import Any, Dict, Optional

from channels.db import database_sync_to_async
from channels.generic.websocket import AsyncJsonWebsocketConsumer

from rest_live import PermissionLambda, __model_to_listeners


def get_group_name(model_label, value, key_prop) -> str:
    return f"RESOURCE-{model_label}-{key_prop}-{value}"


@database_sync_to_async
def does_have_permission(
    user, model_label, instance_filter, check: PermissionLambda
) -> bool:
    from django.apps import apps

    model = apps.get_model(model_label)
    instance = model.objects.get(**instance_filter)
    return check(user, instance)


def get_permission_check(model_label, group_key):
    return __model_to_listeners.get(model_label, dict()).get(group_key, (None, None))[1]


class SubscriptionConsumer(AsyncJsonWebsocketConsumer):
    async def connect(self):
        self.user = self.scope.get("user", None)  # noqa
        await self.accept()

    async def notify(self, event):
        content = event["content"]
        group_key = event["group_key"]
        instance_pk = event["instance_pk"]
        model_label = content["model"]

        check: Optional[PermissionLambda] = get_permission_check(model_label, group_key)

        if check is None:
            await self.send_json(event["content"])
            return

        has_permission = await does_have_permission(
            self.user, model_label, {"id": instance_pk}, check
        )
        if has_permission:
            await self.send_json(event["content"])

    async def receive_json(self, content: Dict[str, Any], **kwargs):
        """
        Receive a subscription on this consumer.
        """

        model_label = content.get("model")
        if model_label is None:
            print("[LIVE] No model")
            return
        prop = content.get("property")
        value = content.get("value", None)

        if value is None:
            print(f"[LIVE] No value for prop {prop} found.")
            return

        group_name = get_group_name(model_label, value, prop)
        unsubscribe = content.get("unsubscribe", False)
        if not unsubscribe:
            print(f"[REST-LIVE] got subscription to {group_name}")
            self.groups.append(group_name)
            await self.channel_layer.group_add(group_name, self.channel_name)
        else:
            print(f"[REST-LIVE] got unsubscribe for {group_name}")
            self.groups.remove(group_name)
            if group_name not in self.groups:
                await self.channel_layer.group_discard(group_name, self.channel_name)



from typing import Callable, Dict, Optional, Tuple, Type

from django.conf import settings
from django.db.models import Model
from rest_framework import serializers


default_app_config = "rest_live.apps.RestLiveConfig"

User = settings.AUTH_USER_MODEL

PermissionLambda = Callable[[User, Model], bool]
SerializerClass = Type[serializers.Serializer]
ListenerEntry = Dict[str, Tuple[SerializerClass, Optional[PermissionLambda]]]

__model_to_listeners: Dict[str, ListenerEntry] = dict()



from typing import Optional

from django.db.models import Model
from rest_framework import serializers

from rest_live import PermissionLambda, __model_to_listeners


def __register_subscription(
    cls, group_key, check_permission: Optional[PermissionLambda]
):
    model: Model = cls.Meta.model
    label = model._meta.label  # noqa
    if label not in __model_to_listeners:
        __model_to_listeners[label] = dict()
    __model_to_listeners[label][group_key] = (cls, check_permission)


def subscribable(group_key: str = "pk", check_permission: PermissionLambda = None):
    def decorator(cls):
        if issubclass(cls, serializers.ModelSerializer):
            __register_subscription(cls, group_key, check_permission)
        return cls

    return decorator



# Generated by Django 3.1 on 2020-08-29 20:32

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="List",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=64)),
            ],
        ),
        migrations.CreateModel(
            name="Todo",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("text", models.CharField(max_length=140)),
                ("done", models.BooleanField(default=False)),
                (
                    "list",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, to="test_app.list"
                    ),
                ),
            ],
        ),
    ]



import os
import sys

from setuptools import find_packages, setup
from setuptools.command.install import install


VERSION = "0.2.1"
DESCRIPTION = open("README.md", encoding="utf-8").read()


class VerifyVersionCommand(install):
    """Custom command to verify that the git tag matches our version"""

    description = "verify that the git tag matches our version"

    def run(self):
        tag = os.getenv("CIRCLE_TAG")

        if tag != VERSION:
            info = "Git tag: {0} does not match the version of this app: {1}".format(
                tag, VERSION
            )
            sys.exit(info)


setup(
    name="django-rest-live",
    version=VERSION,
    packages=find_packages(exclude=["tests", "test_app"]),
    url="https://github.com/pennlabs/django-rest-live",
    project_urls={
        "Changelog": (
            "https://github.com/pennlabs/django-rest-live/blob/master/CHANGELOG.md"
        )
    },
    license="MIT",
    author="Penn Labs",
    author_email="admin@pennlabs.org",
    description="Subscriptions for Django REST Framework over Websockets.",
    long_description=DESCRIPTION,
    long_description_content_type="text/markdown",
    install_requires=["django>=3.0.0", "channels>=2.0.0", "djangorestframework"],
    classifiers=[
        "Framework :: Django",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    python_requires=">=3.6",
    cmdclass={"verify": VerifyVersionCommand},
)



from django.apps import AppConfig
from django.db.models.signals import post_delete, post_save

from rest_live.signals import ondelete_callback, onsave_callback


class RestLiveConfig(AppConfig):
    name = "rest_live"

    def ready(self):
        post_save.connect(onsave_callback)
        post_delete.connect(ondelete_callback)
